from datetime import datetime, timedelta
import logging
import base64
from typing import Any

import httpx

from obelisk.exceptions import AuthenticationError, ObeliskError
from obelisk.strategies.retry import RetryStrategy, NoRetryStrategy
from obelisk.types import ObeliskKind


class BaseClient:
    """
    Base class handling Obelisk auth and doing the core HTTP communication.
    Only exists in asynchronous variety, as it is not usually directly useful for user code.
    """

    _client: str = ""
    _secret: str = ""

    _token: str | None = None
    """Current authentication token"""
    _token_expires: datetime | None = None
    """Deadline after which token is no longer useable"""

    grace_period: timedelta = timedelta(seconds=10)
    """Controls how much before the expiration deadline a token will be refreshed."""
    retry_strategy: RetryStrategy
    kind: ObeliskKind

    log: logging.Logger

    def __init__(
        self,
        client: str,
        secret: str,
        retry_strategy: RetryStrategy = NoRetryStrategy(),  # noqa: B008   # This is fine to bew shared
        kind: ObeliskKind = ObeliskKind.CLASSIC,
    ) -> None:
        self._client = client
        self._secret = secret
        self.retry_strategy = retry_strategy
        self.kind = kind

        self.log = logging.getLogger("obelisk")

    async def _get_token(self) -> None:
        auth_string = str(
            base64.b64encode(f"{self._client}:{self._secret}".encode()), "utf-8"
        )
        headers = {
            "Authorization": f"Basic {auth_string}",
            "Content-Type": (
                "application/json"
                if self.kind.use_json_auth
                else "application/x-www-form-urlencoded"
            ),
        }
        payload = {"grant_type": "client_credentials"}

        async with httpx.AsyncClient() as client:
            request: httpx.Response | None = None
            response: dict[Any, str] | None = None
            last_error = None
            retry = self.retry_strategy.make()
            while not response:
                try:
                    request = await client.post(
                        self.kind.token_url,
                        json=payload if self.kind.use_json_auth else None,
                        data=payload if not self.kind.use_json_auth else None,
                        headers=headers,
                    )
                    if request:
                        response = request.json()
                    break
                except Exception as e:  # noqa: PERF203 # retry strategy should add delay
                    last_error = e
                    self.log.error(e)

                if not await retry.should_retry():
                    break

            if response is None and last_error is not None:
                raise last_error

            if request is None:
                raise last_error or ObeliskError("Could not create HTTP request")

            if request.status_code != 200 or not response:
                if response and "error" in response:
                    self.log.warning(f"Could not authenticate, {response['error']}")
                raise AuthenticationError

            self._token = response["access_token"]
            self._token_expires = datetime.now() + timedelta(
                seconds=float(response["expires_in"])
            )

    async def _verify_token(self) -> None:
        if (
            self._token is None
            or self._token_expires is None
            or self._token_expires <= (datetime.now() + self.grace_period)
        ):
            try:
                await self._get_token()
                return
            except:  # noqa: E722
                # Retry is handled insdie get_token()
                self.log.info("excepted, Cannot get token!")

    async def http_post(
        self, url: str, data: Any = None, params: dict[str, str] | None = None
    ) -> httpx.Response:
        """
        Send an HTTP POST request to Obelisk,
        with proper auth.

        Possibly refreshes the authentication token and performs backoff as per `retry_strategy`.
        This method is not of stable latency because of these properties.

        No validation is performed on the input data,
        callers are responsible for formatting it in a method Obelisk understands.
        """

        await self._verify_token()

        headers = {
            "Authorization": f"Bearer {self._token}",
            "Content-Type": "application/json",
        }
        if params is None:
            params = {}
        async with httpx.AsyncClient() as client:
            response = None
            retry = self.retry_strategy.make()
            last_error = None
            while not response or await retry.should_retry():
                if response is not None:
                    self.log.debug(f"Retrying, last response: {response.status_code}")

                try:
                    response = await client.post(
                        url,
                        json=data,
                        params={k: v for k, v in params.items() if v is not None},
                        headers=headers,
                    )

                    if response.status_code // 100 == 2:
                        return response
                except Exception as e:
                    self.log.error(e)
                    last_error = e
                    continue

            if not response and last_error:
                raise last_error
            return response

    async def http_get(
        self, url: str, params: dict[str, str] | None = None
    ) -> httpx.Response:
        """
        Send an HTTP GET request to Obelisk,
        with proper auth.

        Possibly refreshes the authentication token and performs backoff as per `retry_strategy`.
        This method is not of stable latency because of these properties.

        No validation is performed on the input data,
        callers are responsible for formatting it in a method Obelisk understands.
        """

        await self._verify_token()

        headers = {
            "Authorization": f"Bearer {self._token}",
            "Content-Type": "application/json",
        }
        if params is None:
            params = {}
        async with httpx.AsyncClient() as client:
            response = None
            retry = self.retry_strategy.make()
            last_error = None
            while not response:
                if response is not None:
                    self.log.debug(f"Retrying, last response: {response.status_code}")

                try:
                    response = await client.get(
                        url,
                        params={k: v for k, v in params.items() if v is not None},
                        headers=headers,
                    )

                    if response.status_code // 100 == 2:
                        return response
                except Exception as e:
                    self.log.error(e)
                    last_error = e

                if not await retry.should_retry():
                    break

            if not response and last_error:
                raise last_error
            if response:
                return response
            raise ObeliskError("Could not complete HTTP operation")



"""
Types specific to Obelisk CORE, including an RSQL filter implementation

To create a filter, look at `Filter`.
Example:

>>> from datetime import datetime
>>> f = (Filter().add_and(
...     Comparison.equal('source', 'test source'),
...     Comparison.is_in('metricType', ['number', 'number[]']),
... ).add_or(
...     Comparison.less('timestamp', datetime.fromtimestamp(1757422128))
... ))
>>> print(f)
(('source'=='test source';'metricType'=in=('number', 'number[]')),'timestamp'<'2025-09-09T14:48:48')
"""

from __future__ import annotations
from abc import ABC
from datetime import datetime
from typing import Any
from collections.abc import Iterable
from enum import Enum


FieldName = str
"""https://obelisk.pages.ilabt.imec.be/obelisk-core/query.html#available-data-point-fields
Field names are not validated at this time, due to the inherent complexity.
"""


class Constraint(ABC):  # noqa: B024 # This is just a marker class
    """
    Constraints are simply groups of `Comparison`,
    such as `And`, or `Or`.

    These constraints always enclose their contents in parentheses,
    to avoid confusing precendence situations in serialised format.

    Comparisons should not be created manually by library consumers,
    instead use `Filter.add_and` or `Filter.add_or`.
    """

    pass


class Comparison:
    """
    Comparisons are the basic items of a `Filter`.
    They consist of a field name, operator, and possibly a value on the right.

    It is strongly suggested you create comparisons by using the staticmethods
    on this class, rather than trying to construct them yourselves.

    When serializing to RSQL format,
    each argument is single quoted as to accept any otherwise reserved characters,
    and serialised using :func:`str`.
    """

    left: FieldName
    right: Any
    op: str

    def __init__(self, left: FieldName, right: Any, op: str):
        self.left = left
        self.right = right
        self.op = op

    def __str__(self) -> str:
        right = self._sstr(self.right)
        if not right.startswith("(") and len(right) > 0:
            right = f"'{right}'"

        return f"'{self.left}'{self.op}{right}"

    @staticmethod
    def _sstr(item: Any) -> str:
        """Smart string conversion"""
        if isinstance(item, datetime):
            return item.isoformat()
        return str(item)

    @staticmethod
    def _iterable_to_group(iter: Iterable[Any]) -> str:
        """Produces a group of the form ("a","b")"""
        return str(tuple([Comparison._sstr(x) for x in iter]))

    @classmethod
    def equal(cls, left: FieldName, right: Any) -> Comparison:
        return Comparison(left, right, "==")

    @classmethod
    def not_equal(cls, left: FieldName, right: Any) -> Comparison:
        return Comparison(left, right, "!=")

    @classmethod
    def less(cls, left: FieldName, right: Any) -> Comparison:
        return Comparison(left, right, "<")

    @classmethod
    def less_equal(cls, left: FieldName, right: Any) -> Comparison:
        return Comparison(left, right, "<=")

    @classmethod
    def greater(cls, left: FieldName, right: Any) -> Comparison:
        return Comparison(left, right, ">")

    @classmethod
    def greater_equal(cls, left: FieldName, right: Any) -> Comparison:
        return Comparison(left, right, ">=")

    @classmethod
    def is_in(cls, left: FieldName, right: Iterable[Any]) -> Comparison:
        return Comparison(left, cls._iterable_to_group(right), "=in=")

    @classmethod
    def is_not_in(cls, left: FieldName, right: Iterable[Any]) -> Comparison:
        return Comparison(left, cls._iterable_to_group(right), "=out=")

    @classmethod
    def null(cls, left: FieldName) -> Comparison:
        return Comparison(left, "", "=null=")

    @classmethod
    def not_null(cls, left: FieldName) -> Comparison:
        return Comparison(left, "", "=notnull=")


Item = Constraint | Comparison


class And(Constraint):
    content: list[Item]

    def __init__(self, *args: Item):
        self.content = list(args)

    def __str__(self) -> str:
        return "(" + ";".join([str(x) for x in self.content]) + ")"


class Or(Constraint):
    content: list[Item]

    def __init__(self, *args: Item):
        self.content = list(args)

    def __str__(self) -> str:
        return "(" + ",".join([str(x) for x in self.content]) + ")"


class Filter:
    """
    Filter is an easier way to programatically create filters for the Obelisk CORE platform.

    We still recommend you familiarise yourself with the [CORE filter documentation](https://obelisk.pages.ilabt.imec.be/obelisk-core/query.html#rsql-format),
    as not everything is typechecked.
    Specifically, the left hand side of any comparison is left unchecked.
    Checking this would be borderline impossible with the optional arguments to some fields,
    and the dot notation for labels.

    As this field is not checked, we also do not check whether the type of left operand
    and right operand make sense in the provided comparison.
    """

    content: Item | None = None

    def __init__(self, content: Constraint | None = None):
        self.content = content

    def __str__(self) -> str:
        return str(self.content)

    def add_and(self, *other: Item) -> Filter:
        """
        Encloses the current filter contents, if any,
        in an :class:`And`, along with any other provided Items.

        >>> str(Filter(Comparison.not_null('labels.username'))
        ... .add_and(Comparison.equal('labels.username', 'stpletin'), Comparison.less_equal('timestamp', '2025-01-12T00:00:00Z')))
        "('labels.username'=notnull=;'labels.username'=='stpletin';'timestamp'<='2025-01-12T00:00:00Z')"
        """
        if self.content is None:
            self.content = And(*other)
        else:
            self.content = And(self.content, *other)
        return self

    def add_or(self, *other: Item) -> Filter:
        """
        Encloses the current filter contents, if any,
        in an :class:`Or`, along with any other provided Items.

        >>> str(Filter(Comparison.not_null('labels.username'))
        ... .add_or(Comparison.equal('labels.username', 'stpletin'), Comparison.less_equal('timestamp', '2025-01-12T00:00:00Z')))
        "('labels.username'=notnull=,'labels.username'=='stpletin','timestamp'<='2025-01-12T00:00:00Z')"
        """
        if self.content is None:
            self.content = Or(*other)
        else:
            self.content = Or(self.content, *other)
        return self


class IngestMode(str, Enum):
    """Whether the ingested datapoints should be streamed, stored, or both (the default)"""
    BOTH = "DEFAULT"
    STREAM = "STREAM_ONLY"
    STORE = "STORE_ONLY"



from enum import Enum
from typing import Any

from pydantic import BaseModel, ConfigDict


class IngestMode(str, Enum):
    """
    Classic Obelisk accepts three ways of submitting new data.
    This integrates with the concept of Streams,
    which is a way to process datapoints as they come in.

    The default submission method is to publish both to long-term storage
    and stream.
    This IngestMode can be changed to change this default.

    Does not apply to HFS
    """

    DEFAULT = "default"
    STREAM_ONLY = "stream_only"
    STORE_ONLY = "store_only"


class TimestampPrecision(str, Enum):
    """
    When ingesting data it is important to specify which precision provided UNIX timestamps are in.
    If a provided timestamp is in seconds,
    but interpreted by Obelisk as milliseconds, it would erroneously be somewhere in the past.
    """

    __choices__ = ("SECONDS", "MILLISECONDS", "MICROSECONDS")

    SECONDS = "seconds"
    MILLISECONDS = "milliseconds"
    MICROSECONDS = "microseconds"


class Datapoint(BaseModel):
    """An Obelisk Classic / HFS datapoint. May contain more or less fields"""

    timestamp: int
    value: Any
    dataset: str | None = None
    metric: str | None = None
    source: str | None = None
    userId: int | None = None
    """This field is only used on HFS, and has a different name in some deployments."""

    model_config = ConfigDict(extra="allow")


class QueryResult(BaseModel):
    """Result of a query"""

    items: list[Datapoint]
    cursor: str | None = None


class ObeliskKind(str, Enum):
    """Defines which variety of Obelisk a Client should use, and provides some URLs and config information."""

    CLASSIC = "classic"
    HFS = "hfs"
    CORE = "core"

    @property
    def token_url(self) -> str:
        match self:
            case ObeliskKind.CLASSIC:
                return "https://obelisk.ilabt.imec.be/api/v3/auth/token"
            case ObeliskKind.HFS:
                return "https://obelisk-hfs.discover.ilabt.imec.be/auth/realms/obelisk-hfs/protocol/openid-connect/token"
            case ObeliskKind.CORE:
                return "https://auth.obelisk.discover.ilabt.imec.be/realms/obelisk/protocol/openid-connect/token"

    @property
    def root_url(self) -> str:
        match self:
            case ObeliskKind.CLASSIC:
                return "https://obelisk.ilabt.imec.be/api/v3"
            case ObeliskKind.HFS:
                return "https://obelisk-hfs.discover.ilabt.imec.be"
            case ObeliskKind.CORE:
                return "https://obelisk.discover.ilabt.imec.be/datasets"

    @property
    def query_url(self) -> str:
        match self:
            case ObeliskKind.CLASSIC:
                return "https://obelisk.ilabt.imec.be/api/v3/data/query/events"
            case ObeliskKind.HFS:
                return "https://obelisk-hfs.discover.ilabt.imec.be/data/query/events"
            case ObeliskKind.CORE:
                raise NotImplementedError()

    @property
    def ingest_url(self) -> str:
        match self:
            case ObeliskKind.CLASSIC:
                return "https://obelisk.ilabt.imec.be/api/v3/data/ingest"
            case ObeliskKind.HFS:
                return "https://obelisk-hfs.discover.ilabt.imec.be/data/ingest"
            case ObeliskKind.CORE:
                raise NotImplementedError()

    @property
    def stream_url(self) -> str | None:
        match self:
            case ObeliskKind.CLASSIC:
                return "https://obelisk.ilabt.imec.be/api/v3/data/streams"
            case ObeliskKind.HFS:
                return None
            case ObeliskKind.CORE:
                raise NotImplementedError()

    @property
    def use_json_auth(self) -> bool:
        match self:
            case ObeliskKind.CLASSIC:
                return True
            case ObeliskKind.HFS:
                return False
            case ObeliskKind.CORE:
                return False



"""
This module provides wrappers for the classes in `obelisk.asynchronous` with a synchronous API.
These hold on to a private event loop and block until a result is available.

Note
----

Please note that the sync module *will not* work in an asynchronous context (i.e. in any method returning a Future, or a Jupyter notebook).
This is because it is internally nothing more than a wrapper over the asynchronous implementation.
Use the asynchronous implementation in these situations.
"""

__all__ = ["Obelisk"]
from .client import Obelisk



"""
Obelisk-py is a client for the Obelisk data platform.
We support both "classic" Obelisk and HFS,
each with a synchronous and async API.
We also support Obelisk CORE, in async only for now.

The PyPi package name is `obelisk-py`, the Python module is called `obelisk`.

Your starting point will be one of the Obelisk instances in `.sync` or `.asynchronous` depending on your preferred API.
The Obelisk classes in these modules both implement the same interface,
but the asynchronous implementation returns Coroutines.

For Obelisk CORE, look at `obelisk.asynchronous.core`, or the below example as your starting point.

## Error handling

Obelisk-py comes with robust retry logic to handle any errors that may come up.
Issues like timeouts, temporary server errors or even DNS issues are fairly common, and handling them properly is important.
Each Client accepts a retry strategy of type `.strategies.retry.RetryStrategy`.
Several predefined strategies are available in `.strategies.retry`.

## Quick Start

Using CORE
```py
from obelisk.asynchronous.core import Client, QueryParams
from obelisk.types.core import Filter, Comparison
import os
import asyncio

client_id = os.getenv('CLIENT_ID')
client_secret = os.getenv('CLIENT_SECRET')

# You may want to specify a retry strategy
client = Client(
    client=client_id,
    secret=client_secret
)

query = QueryParams(
    dataset="some-dataset",
    fields=["metric","labels","value","timestamp"],
    dataType="number",
    filter_=Filter().add_and(
        Comparison.equal("metric", "heart_rate::number"),
        Comparison.greater("timestamp", 42069180)
    )

data = asyncio.get_event_loop().run_until_complete(client.query(query))
```

Using Classic or HFS, synchronously (async is analogous)
```py
from obelisk.sync import Client
from obelisk.types import ObeliskKind
import os

client_id = os.getenv('CLIENT_ID')
client_secret = os.getenv('CLIENT_SECRET')

# You may want to specify a retry strategy
client = Client(
    client=client_id,
    secret=client_secret,
    kind=ObeliskKind.CLASSIC # or HFS, as you wish
)

data = client.query(
    datasets=["some-dataset"],
    metrics=["heart-rate::number"],
    from_timestamp=42069180,
    filter_={
        "source": {
            "_startsWith": "user123"
        }
    }
)
```

## Changelog

See [here](https://github.com/predict-idlab/obelisk-python/blob/main/CHANGELOG.rst)
"""



from obelisk.sync import Obelisk

client_id = "682c6c46604b3b3be35429df"
client_secret = "7136832d-01be-456a-a1fe-25c7f9e130c5"


def test_demo_igent_fetch():
    consumer = Obelisk(client=client_id, secret=client_secret)
    result = consumer.fetch_single_chunk(
        datasets=["612f6c39cbceda0ea9753d95"],
        metrics=["org.dyamand.types.common.Temperature::number"],
        from_timestamp=1740924034000,
        to_timestamp=1741100614258,
        limit=2,
    )

    assert len(result.items) == 2


def test_two_instances():
    consumer_one = Obelisk(client=client_id, secret=client_secret)
    consumer_two = Obelisk(client=client_id, secret=client_secret)
    result_one = consumer_one.fetch_single_chunk(
        datasets=["612f6c39cbceda0ea9753d95"],
        metrics=["org.dyamand.types.common.Temperature::number"],
        from_timestamp=1740924034000,
        to_timestamp=1741100614258,
        limit=2,
    )
    result_two = consumer_two.fetch_single_chunk(
        datasets=["612f6c39cbceda0ea9753d95"],
        metrics=["org.dyamand.types.common.Temperature::number"],
        from_timestamp=1740924034000,
        to_timestamp=1741100614258,
        limit=2,
    )
    assert len(result_one.items) == 2
    assert len(result_two.items) == 2



from obelisk.asynchronous.core import QueryParams
from obelisk.types.core import Filter, Comparison


def test_query_param_serialize():
    q = QueryParams(
        dataset="83989232",
        filter_="(metric=='smartphone.application::string')",
        dataType="string",
    )
    dump = q.to_dict()
    assert "filter" in dump


def test_comma_separate_string():
    data = QueryParams(
        dataset="dummy",
        groupBy=["dataset", "timestamp"],
        fields=["timestamp", "labels", "value"],
        orderBy=["timestamp", "value"],
        dataType="string",
    ).to_dict()

    assert data["fields"] == "timestamp,labels,value"
    assert data["groupBy"] == "dataset,timestamp"
    assert data["orderBy"] == "timestamp,value"



def test_filter_serialize():
    filter_ = Filter().add_and(Comparison.equal("metric", "smartphone.acceleration::number[]"), Comparison.greater_equal("timestamp", 123456))
    q = QueryParams(
        dataset="83989232",
        filter_=filter_,
        dataType="number[]",
    )
    dump = q.to_dict()
    assert "filter" in dump
